
CREATE TABLE dict (
       word                 VARCHAR2(32) NOT NULL,
       url_id               NUMBER(12) NOT NULL,
       intag                NUMBER(12) NOT NULL
)
       PCTFREE 5
       PCTUSED 95
;


ALTER TABLE dict
       ADD  ( PRIMARY KEY (word, url_id) ) ;


CREATE TABLE robots (
       hostinfo             VARCHAR2(127) NOT NULL,
       path                 VARCHAR2(127) NOT NULL
);


CREATE TABLE stopword (
       word                 VARCHAR(64) NOT NULL
,
       lang                 VARCHAR(32) NOT NULL
);


ALTER TABLE stopword
       ADD  ( PRIMARY KEY (word, lang) ) ;


CREATE TABLE url (
       rec_id               NUMBER(12) NOT NULL,
       status               NUMBER(6) DEFAULT 0 NOT NULL,
       docsize              NUMBER(12) NULL,
       next_index_time      NUMBER(12) NULL,
       last_mod_time        NUMBER(12) NULL,
       referrer             NUMBER(12) DEFAULT 0 NOT NULL,
       hops                 NUMBER(6) DEFAULT 0 NOT NULL,
       crc32                NUMBER(12) NULL,
       seed                 NUMBER(6) DEFAULT 0 NOT NULL,
       bad_since_time       NUMBER(12) NULL,
       site_id		    NUMBER(12),
       server_id	    NUMBER(12),
       pop_rank	  	    FLOAT DEFAULT 0.0 NOT NULL,
       url                  VARCHAR2(128) NOT NULL
)
       PCTFREE 5
       PCTUSED 95
;

CREATE UNIQUE INDEX XAK1url ON url
(
       url                            ASC
);

CREATE INDEX XIE1url ON url
(
       crc32                          ASC
);

CREATE INDEX rec_id_ind ON url
(
       rec_id			
);

CREATE INDEX next_index_time_ind ON url
(
       	next_index_time
);


CREATE SEQUENCE next_url_id
   INCREMENT BY 1
   START WITH 1;


CREATE TABLE urlinfo (
       url_id NUMBER(38) NOT NULL,
       sname  VARCHAR2(64) NOT NULL,
       sval   VARCHAR2(1024)
);

CREATE INDEX urlinfo_id_ind ON urlinfo
(
       url_id			
);


CREATE TABLE server (
	rec_id          NUMBER          DEFAULT 0 NOT NULL, 
	active		NUMBER		DEFAULT 0 NOT NULL,
	url		VARCHAR2(128)	DEFAULT '' NOT NULL,
	alias           VARCHAR2(128)   DEFAULT '' NOT NULL,
	period		NUMBER		DEFAULT 604800 NOT NULL,
	tag		VARCHAR2(16)	DEFAULT '' NOT NULL,
	category	VARCHAR2(16)	DEFAULT '' NOT NULL,
	charset		VARCHAR2(40)	DEFAULT '' NOT NULL,
	lang		VARCHAR2(32)	DEFAULT '' NOT NULL,
	basic_auth	VARCHAR2(64)	DEFAULT '' NOT NULL,
	proxy		VARCHAR2(64)	DEFAULT '' NOT NULL,
	proxy_port	NUMBER		DEFAULT 3128 NOT NULL,
	proxy_auth	VARCHAR2(64)	DEFAULT '' NOT NULL,
	maxhops		NUMBER		    DEFAULT 9999 NOT NULL,
	gindex		NUMBER		    DEFAULT 1 NOT NULL,
	follow		NUMBER		    DEFAULT 1 NOT NULL,
	deletebad	NUMBER		    DEFAULT 0 NOT NULL,
	use_robots	NUMBER		    DEFAULT 1 NOT NULL,
	use_clones	NUMBER		    DEFAULT 1 NOT NULL,
	max_net_errors	NUMBER		    DEFAULT 16 NOT NULL,
	net_delay_time	NUMBER		    DEFAULT 86400 NOT NULL,
	read_timeout	NUMBER		    DEFAULT 90 NOT NULL,
	command		CHAR(1)		DEFAULT 'S' NOT NULL,
	match_type	NUMBER				default 17,
	ordre		NUMBER		not null,
	parent		NUMBER		DEFAULT 0 not null,
	weight		float		DEFAULT 1 not null,
	pop_weight	float		DEFAULT 0 not null
);

CREATE SEQUENCE server_rec_id_seq INCREMENT BY 1 START WITH 1;

ALTER TABLE server ADD CONSTRAINT server_pk PRIMARY KEY (rec_id);

CREATE UNIQUE INDEX srv_url ON server (url); 

CREATE TABLE links (
	ot	 NUMBER not null,
	k	 NUMBER not null,
	weight   FLOAT DEFAULT 0.0 not null
);

CREATE UNIQUE INDEX links_links ON links (ot, k);
CREATE INDEX links_ot ON links (ot);
CREATE INDEX links_k ON links (k);

